import os, re, subprocess, pytest
from .test_recipe import change_test_dir, run, verify_output

def test_conditional_skips():
    os.system("rm -fr test_conditional_skips[1234].tmp")
    print("===== all three files touched =====")
    retcode, output = run("stimela -b native run test_conditional_skips.yml")
    assert retcode == 0
    print(output)
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips1.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips2.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips3.tmp")

    print("===== 2 and 3 touched =====")
    os.system("touch test_conditional_skips4.tmp")
    retcode, output = run("stimela -b native run test_conditional_skips.yml")
    assert retcode == 0
    print(output)
    assert not verify_output(output, "---INVOKING---", "touch test_conditional_skips1.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips2.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips3.tmp")

    print("===== only 3 touched =====")
    retcode, output = run("stimela -b native run test_conditional_skips.yml")
    assert retcode == 0
    print(output)
    assert not verify_output(output, "---INVOKING---", "touch test_conditional_skips1.tmp")
    assert not verify_output(output, "---INVOKING---", "touch test_conditional_skips2.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips3.tmp")

    print("===== 2 and 3 touched =====")
    retcode, output = run("stimela -b native run test_conditional_skips.yml -f")
    assert retcode == 0
    print(output)
    assert not verify_output(output, "---INVOKING---", "touch test_conditional_skips1.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips2.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips3.tmp")

    print("===== 1 and 3 touched =====")
    retcode, output = run("stimela -b native run test_conditional_skips.yml -F")
    assert retcode == 0
    print(output)
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips1.tmp")
    assert not verify_output(output, "---INVOKING---", "touch test_conditional_skips2.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips3.tmp")

    print("===== all three files touched =====")
    retcode, output = run("stimela -b native run test_conditional_skips.yml -f -F")
    assert retcode == 0
    print(output)
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips1.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips2.tmp")
    assert verify_output(output, "---INVOKING---", "touch test_conditional_skips3.tmp")

    print("===== cleaning up =====")
    os.system("rm -fr test_conditional_skips[1234].tmp")



import argparse
import yaml

import os
import sys
from uuid import uuid4


def create_parser():
    p = argparse.ArgumentParser()
    p.add_argument("-j", "--job-name", default="job", help="Update the job name")
    p.add_argument(
        "-n",
        "--namespace",
        default="rarg-test-compute",
        help="The kubernetes namespace in which the job will be run",
    )
    p.add_argument(
        "-i",
        "--image",
        default="ghcr.io/dask/dask",
        help="Container image run by job, scheduler and worker pods",
    )
    p.add_argument(
        "-s",
        "--service-account",
        default="compute-runner",
        help="The kubernetes service account which will run the job",
    )
    p.add_argument(
        "--pull-policy",
        default="Always",
        help="imagePullPolicy setting for pods",
    )
    p.add_argument(
        "-f",
        "--mount-file",
        nargs="*",
        help=(
            "Configuration files that will be mounted into the pod. "
            "Can be of the form /path/to/file.yaml or "
            "/host/file.yaml:/container/file.yaml. "
            "If the first form is selected, file.yaml will be "
            "mounted at /mnt/file.yaml"
        ),
    )
    p.add_argument(
        "-v",
        "--volume",
        nargs="*",
        help=(
            "PersistentVolumeClaims that will be mounted into the pod."
            "Should be of the form pvc-name:mount_point"
        ),
    )
    p.add_argument(
        "-w", "--nworkers", default=1, type=int, help="Number of dask workers to launch"
    )
    p.add_argument(
        "-cr",
        "--cpu-request",
        default=1,
        type=float,
        help="Number of cpu's to request for the worker pods",
    )
    p.add_argument(
        "-mr",
        "--mem-request",
        default="1Gi",
        help="Amount of memory to request for the worker pods",
    )
    p.add_argument(
        "-cl",
        "--cpu-limit",
        default=1,
        type=float,
        help="cpu limit for the worker pods",
    )
    p.add_argument(
        "-ml", "--mem-limit", default="1Gi", help="memory limit for the worker pods"
    )

    return p


def split_args(args):
    try:
        i = args.index("--")
    except ValueError:
        return args, []
    else:
        return args[:i], args[i + 1 :]


def daskjob_template(args):
    labels = dict(args.labels)

    env_var = [{"name": k, "value": v} for k, v in args.environment_variables.items()]

    return {
        "apiVersion": "kubernetes.dask.org/v1",
        "kind": "DaskJob",
        "metadata": {"name": args.job_name, "namespace": args.namespace, "labels": labels },
        "spec": {
            "cluster": {
                "spec": {
                    "scheduler": {
                        "service": {
                            "ports": [
                                {
                                    "name": "tcp-comm",
                                    "port": 8786,
                                    "protocol": "TCP",
                                    "targetPort": "tcp-comm",
                                },
                                {
                                    "name": "http-dashboard",
                                    "port": 8787,
                                    "protocol": "TCP",
                                    "targetPort": "http-dashboard",
                                },
                            ],
                            "selector": {
                                "dask.org/cluster-name": args.job_name,
                                "dask.org/component": "scheduler",
                            },
                            "type": "ClusterIP",
                        },
                        "metadata": { "labels": labels },
                        "spec": {
                            "containers": [
                                {
                                    "args": ["dask-scheduler"],
                                    "env": [
                                        {
                                            "name": "SCHEDULER_ENV",
                                            "value": "hello-world",
                                        },
                                        *env_var
                                    ],
                                    "image": args.image,
                                    "imagePullPolicy": args.pull_policy,
                                    "livenessProbe": {
                                        "httpGet": {
                                            "path": "/health",
                                            "port": "http-dashboard",
                                        },
                                        "initialDelaySeconds": 15,
                                        "periodSeconds": 20,
                                    },
                                    "name": "scheduler",
                                    "ports": [
                                        {
                                            "containerPort": 8786,
                                            "name": "tcp-comm",
                                            "protocol": "TCP",
                                        },
                                        {
                                            "containerPort": 8787,
                                            "name": "http-dashboard",
                                            "protocol": "TCP",
                                        },
                                    ],
                                    "readinessProbe": {
                                        "httpGet": {
                                            "path": "/health",
                                            "port": "http-dashboard",
                                        },
                                        "initialDelaySeconds": 5,
                                        "periodSeconds": 10,
                                    },
                                }
                            ],
                        },
                    },
                    "worker": {
                        "replicas": args.nworkers,
                        "metadata": { "labels": labels },
                        "spec": {
                            "containers": [
                                {
                                    "args": [
                                        "dask-worker",
                                        "--name",
                                        "$(DASK_WORKER_NAME)",
                                        "--nworkers", '1',
                                        "--nthreads", str(args.threads_per_worker),
                                        "--memory-limit", str(args.memory_limit),
                                        "$(DASK_SCHEDULER_ADDRESS)",
                                    ],
                                    "env": [
                                        {
                                            "name": "WORKER_ENV",
                                            "value": "hello-world"
                                        },
                                        *env_var
                                    ],
                                    "image": args.image,
                                    "imagePullPolicy": args.pull_policy,
                                    "name": "worker",
                                }
                            ],
                        },
                    },
                }
            },
            "job": {
                "metadata": { "labels": labels},
                "spec": {
                    "containers": [
                        {
                            "image": args.image,
                            "imagePullPolicy": args.pull_policy,
                            "name": "job",
                        }
                    ],
                }
            },
        },
    }


def parse_mount_file(mount_file):
    bits = mount_file.split(":")

    if len(bits) == 2:
        host, container = bits
        _, key = os.path.split(host)
    elif len(bits) == 1:
        host = bits[0]
        _, key = os.path.split(host)
        container = f"/mnt/{key}"
    else:
        raise ValueError(
            f"mount_file should be of the form "
            f"'/mnt/filename' or "
            f"'/host/path/to/filename:/container/path/to/filename' "
            f"Got {mount_file}"
        )

    return host, container, key


def render(args):
    d = daskjob_template(args)
    segments = [d]

    job_spec = d["spec"]["job"]["spec"]
    worker_spec = d["spec"]["cluster"]["spec"]["worker"]["spec"]
    scheduler_spec = d["spec"]["cluster"]["spec"]["scheduler"]["spec"]

    if args.service_account:
        for entry in [job_spec, worker_spec, scheduler_spec]:
            entry["serviceAccountName"] = args.service_account
            entry["automountServiceAccountToken"] = True

    if args.mount_file:
        for i, mount in enumerate(args.mount_file):
            host_path, container_path, mount_key = parse_mount_file(mount)

            with open(host_path, "r") as f:
                payload = f.read()

            volume_name = f"config-map-volume-{i}"
            mount_name = f"file-mount-{uuid4().hex[:6]}"

            segments.append(
                {
                    "apiVersion": "v1",
                    "data": {mount_key: payload},
                    "kind": "ConfigMap",
                    "metadata": {
                        "name": mount_name,
                        "namespace": args.namespace,
                    },
                }
            )

            for entry in [job_spec, worker_spec, scheduler_spec]:
                volumes = entry.setdefault("volumes", [])

                for container in entry["containers"]:
                    volume_mounts = container.setdefault("volumeMounts", [])

                    volume_mounts.append(
                        {
                            "name": volume_name,
                            "mountPath": container_path,
                            "subPath": mount_key,
                        }
                    )
                    volumes.append(
                        {
                            "name": volume_name,
                            "configMap": {"name": mount_name},
                        }
                    )

    return segments


if __name__ == "__main__":
    args, cmdline = split_args(sys.argv[1:])
    args = create_parser().parse_args(args)
    args.cmdline = cmdline
    yaml.dump_all(render(args), sys.stdout, default_flow_style=False)



import os
import logging
from shutil import which
from dataclasses import dataclass
from omegaconf import OmegaConf
from typing import Dict, List, Any, Optional
from scabha.basetypes import EmptyDictDefault

from stimela.exceptions import BackendError

# path to default srun binary
_default_srun_path = None

@dataclass
class SlurmOptions(object):
    enable: bool = False                            # enables passing off jobs to slurm via srun
    srun_path: Optional[str] = None                 # path to srun executable
    srun_opts: Dict[str, str] = EmptyDictDefault()  # extra options passed to srun. "--" prepended, and "_" replaced by "-"
    srun_opts_build: Optional[Dict[str, str]] = None  # extra options passed to srun for build commands. If None, use srun_opts
    build_local: bool = True                        # if True, images will be built locally (i.e. on the head node) even when slurm is enabled
    
    # ## disabling this for now
    # # these will be checked for
    # required_mem_opts: Optional[List[str]] = ListDefault("mem", "mem-per-cpu", "mem-per-gpu")
    # # this will be applied if the required above are missing
    # default_mem_opt: str = "8GB"
    

    def get_executable(self):
        global _default_srun_path
        if self.srun_path is None:
            if _default_srun_path is None:
                _default_srun_path = which("srun")
                if not _default_srun_path:
                    _default_srun_path = False
            if _default_srun_path is False:
                raise BackendError(f"slurm 'srun' binary not found")
            return _default_srun_path
        else:
            if not os.access(self.srun_path, os.X_OK):
                raise BackendError(f"slurm.srun_path '{self.srun}' is not an executable")
            return self.srun
        
    def _wrap(self, srun_opts: Dict[str, Any], args: List[str],  log_args: List[str], 
                ephem_binds: Dict[str, str] = {}, fqname: Optional[str]=None) -> List[str]:
        output_args = [self.get_executable()]

        # reverse fqname to make job name (more informative that way)
        if fqname is not None:
            output_args += ["-J", '.'.join(fqname.split('.')[::-1])]

        # add all base options that have been specified
        for name, value in srun_opts.items():
            output_args += ["--" + name, value]

        # use wrapper script if ephemeral bindings are required
        if ephem_binds:
            output_args.append(os.path.join(os.path.dirname(__file__), 'slurm_runner.sh'))
            for name, basedir in ephem_binds.items():
                output_args.append(f"{basedir}::{name}") 
            output_args.append("--")

        return output_args + args, output_args + log_args

    def wrap_run_command(self, args: List[str], log_args: List[str], 
                            ephem_binds: Dict[str, str] = {}, fqname: Optional[str]=None, log: Optional[logging.Logger]=None) -> List[str]:
        return self._wrap(self.srun_opts, args, log_args, ephem_binds=ephem_binds, fqname=fqname)        

    def wrap_build_command(self, args: List[str], fqname: Optional[str]=None, log: Optional[logging.Logger]=None) -> List[str]:
        if self.build_local:
            return args
        return self._wrap(self.srun_opts_build if self.srun_opts_build is not None else self.srun_opts, args,
                          log_args=args, fqname=fqname)
    
    def validate(self, log: logging.Logger):
        pass
        # if self.required_mem_opts:
        #     if not set(self.srun_opts.keys()).intersection(self.required_mem_opts):
        #         self.srun_opts['mem'] = self.default_mem_opt

SlurmOptionsSchema = OmegaConf.structured(SlurmOptions)



import shlex
import logging
from typing import Optional, Any, Union, Dict
from dataclasses import dataclass

from stimela.exceptions import CabValidationError
from stimela.kitchen.cab import Cab
from scabha.cab_utils import CAB_OUTPUT_PREFIX
from stimela.kitchen import wranglers
from scabha.substitutions import substitutions_from

from . import _CallableFlavour, _BaseFlavour
import stimela

@dataclass
class BinaryFlavour(_BaseFlavour):
    """
    Represents a cab flavour that is a command run via the shell
    """
    kind: str = "binary"

    def get_arguments(self, cab: Cab, params: Dict[str, Any], subst: Dict[str, Any], 
                      virtual_env: Optional[str]=None, check_executable: bool = True, log: Optional[logging.Logger] = None):

        # build command line from parameters
        args = cab.build_command_line(params, subst, virtual_env=virtual_env, check_executable=check_executable)

        # prepend virtual env invocation, if asked
        if virtual_env:
            venv_args = f". {virtual_env}/bin/activate && "
            args = ["/bin/bash", "-c", venv_args + " ".join(shlex.quote(arg) for arg in args)]

        return args, args
    
    def get_image_name(self, cab: Cab, backend: 'stimela.backend.StimelaBackendOptions'):
        from stimela.backends import resolve_image_name
        from stimela import CONFIG
        return resolve_image_name(backend, cab.image or CONFIG.images['default-python'])




from stimela import logger, LOG_FILE, BASE, utils

from stimela.backends import docker, singularity, podman

def make_parser(subparsers):
    parser = subparsers.add_parser("ps", help='List all running stimela processes')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
        help="Clear logfile that keeps track of stimela processes. This doesn't do anything ot the processes themselves.")

    parser.set_defaults(func=ps)


def ps(args, conf):
    pass

    # log = logger.StimelaLogger(LOG_FILE)
    # log.display('processes')
    # if args.clear:
    #     log.clear('processes')
    #     log.write()




from __future__ import annotations
from dataclasses import field, dataclass
from collections import OrderedDict
from typing import List, Union, get_args, get_origin, Any
import os.path
import re
from .exceptions import UnsetError
from itertools import zip_longest
from typeguard import (
    check_type, TypeCheckError, TypeCheckerCallable, TypeCheckMemo, checker_lookup_functions
)
from inspect import isclass
import uritools
from pathlib import Path


def EmptyDictDefault():
    return field(default_factory=lambda:OrderedDict())

def EmptyListDefault():
    return field(default_factory=lambda:[])

def ListDefault(*args):
    return field(default_factory=lambda:list(args))

def DictDefault(**kw):
    return field(default_factory=lambda:dict(**kw))

def EmptyClassDefault(obj):
    return field(default_factory=obj) 


@dataclass
class Unresolved(object):
    value: str = ""
    errors: List[Exception] = EmptyListDefault()

    def __post_init__(self):
        if not self.value:
            self.value = "; ".join(map(str, self.errors))
        if not self.errors:
            self.errors = [UnsetError(f"'{self.value}' undefined")]
        # prevent {}-substitutions on Unresolved message
        self.value = self.value.replace("{", "{{").replace("}", "}}")

    def __str__(self):
        return f"Unresolved({self.value})"

class UNSET(Unresolved):
    """Marks unset values in formulas"""
    pass

class Placeholder(Unresolved):
    """Marks placeholder values that are guaranteed to resolve later, such as for-loop iterants"""
    pass

class SkippedOutput(Unresolved):
    """Marks invalid outputs of skipped steps"""
    def __str__(self):
        return f"Skipped({self.value})"

class URI(str):
    def __init__(self, value):

        uri_components = uritools.urisplit(value)
        self.scheme = uri_components.scheme or "file"  # Protocol.
        self.authority = uri_components.authority
        self.query = uri_components.query
        self.fragment = uri_components.fragment

        # NOTE(JSKenyon): We assume that remote URIs are properly formed and
        # absolute i.e. we do not reason about relative paths for e.g. s3. The
        # following attempts to express paths relative to the cwd but will
        # prefer absolute paths when inputs are outside the cwd. This can be
        # changed when stimela's minimum Python >= 3.12 by using the newly
        # added `walk_up` option.
        if self.scheme == "file":
            cwd = Path.cwd().absolute()
            abs_path = Path(uri_components.path).expanduser().resolve()
            self.abs_path = str(abs_path)
            try:
                self.path = str(abs_path.relative_to(cwd))
            except ValueError as e:
                if "is not in the subpath" in str(e):
                    self.path = self.abs_path
                else:
                    raise e
        else:
            self.path = self.abs_path = uri_components.path

        self.full_uri = uritools.uricompose(
            scheme=self.scheme,
            authority=self.authority,
            path=self.abs_path,
            query=self.query,
            fragment=self.fragment
        )

        self.remote = self.scheme != "file"

    def __str__(self):
        return self.full_uri if self.remote else self.path

    def __repr__(self):
        return self.full_uri


class File(URI):
    @property
    def NAME(self):
        return File(os.path.basename(self))

    @property
    def PATH(self):
        return File(os.path.abspath(self))

    @property
    def DIR(self):
        return File(os.path.dirname(self))

    @property
    def BASEPATH(self):
        return File(os.path.splitext(self)[0])

    @property
    def BASENAME(self):
        return File(os.path.splitext(self.NAME)[0])

    @property
    def EXT(self):
        return os.path.splitext(self)[1]

    @property
    def EXISTS(self):
        return os.path.exists(self)

class Directory(File):
    pass

class MS(Directory):
    pass

FILE_TYPES = (File, MS, Directory, URI)

def is_file_type(dtype):
    return any(dtype == t for t in FILE_TYPES)

def is_file_list_type(dtype):
    return any(dtype == List[t] for t in FILE_TYPES)


def check_filelike(value: Any, origin_type: Any, args: tuple[Any, ...], memo: TypeCheckMemo) -> None:
    """Custom checker for filelike objects. Currently checks for strings."""
    if not isinstance(value, str):
        raise TypeCheckError(f'{value} is not compatible with URI or its subclasses.')


def filelike_lookup(origin_type: Any, args: tuple[Any, ...], extras: tuple[Any, ...]) -> TypeCheckerCallable | None:
    """Lookup the custom checker for filelike objects."""
    if isclass(origin_type) and issubclass(origin_type, URI):
        return check_filelike

    return None

checker_lookup_functions.append(filelike_lookup)  # Register custom type checker.

def get_filelikes(dtype, value, filelikes=None):
    """Recursively recover all filelike elements from a composite dtype."""

    filelikes = set() if filelikes is None else filelikes

    if value is UNSET or isinstance(value, Unresolved):
      return []
      
    origin = get_origin(dtype)
    args = get_args(dtype)

    if origin:  # Implies composition.

        if origin is dict:

            # No further work required for empty collections.
            if len(value) == 0:
                return filelikes

            k_dtype, v_dtype = args

            for k, v in value.items():
                filelikes = get_filelikes(k_dtype, k, filelikes)
                filelikes = get_filelikes(v_dtype, v, filelikes)

        elif origin in (tuple, list, set):

            # No further work required for empty collections.
            if len(value) == 0:
                return filelikes

            # This is a special case for tuples of arbitrary
            # length i.e. list-like behaviour. We can simply 
            # strip out the Ellipsis.
            args = tuple([arg for arg in args if arg != ...])

            for dt, v in zip_longest(args, value, fillvalue=args[0]):
                filelikes = get_filelikes(dt, v, filelikes)

        elif origin is Union:

            for dt in args:
                try:
                    check_type(value, dt)
                except TypeCheckError:  # Value doesn't match dtype - incorrect branch.
                    continue
                filelikes = get_filelikes(dt, value, filelikes)

        else:
            raise ValueError(f"Failed to traverse {dtype} dtype when looking for files.")

    else:
        if is_file_type(dtype):
            filelikes.add(value)

    return filelikes



from stimela import logger, LOG_FILE, BASE, utils
from stimela.backends import docker, singularity, podman


def make_parser(subparsers):
    parser = subparsers.add_parser("containers", help='List all active stimela containers.')

    add = parser.add_argument

    add("-c", "--clear", action="store_true",
        help="Clear the log file that keeps track of stimela containers. This doesn't do anything to the containers.")

    parser.set_defaults(func=containers)


def containers(args, conf):
    log = logger.StimelaLogger(LOG_FILE)
    log.display('containers')
    if args.clear:
        log.clear('containers')
        log.write()




import os, re, subprocess, pytest


# Change into directory where test_recipy.py lives
# As suggested by https://stackoverflow.com/questions/62044541/change-pytest-working-directory-to-test-case-directory
@pytest.fixture(autouse=True)
def change_test_dir(request, monkeypatch):
    monkeypatch.chdir(request.fspath.dirname)


def callable_function(a: int, b: str):
    print(f"callable_function({a},'{b}')")


def run(command):
    """Runs command, returns tuple of exit code, output"""
    print(f"running: {command}")
    try:
        return 0, subprocess.check_output(command, shell=True).strip().decode()
    except subprocess.CalledProcessError as exc:
        return exc.returncode, exc.output.strip().decode()

def verify_output(output, *regexes):
    """Returns true if the output contains lines matching the regexes in sequence (possibly with other lines in between)"""
    regexes = list(regexes[::-1])
    for line in output.split("\n"):
        if regexes and re.search(regexes[-1], line):
            regexes.pop()
    if regexes:
        print("Error, the following regexes did not match the output:")
        for regex in regexes:
            print(f"  {regex}")
        return False
    return True


def test_test_aliasing():
    print("===== expecting an error since required parameters are missing =====")
    retcode, _ = run(f"stimela -v -b native exec test_aliasing.yml")
    assert retcode != 0 

    print("===== expecting no errors now =====")
    retcode, output = run("stimela -v doc test_aliasing.yml")

    print("===== expecting no errors now =====")
    retcode, output = run("stimela -v -b native exec test_aliasing.yml a=1 s3.a=1 s4.a=1 e=e f=f")
    assert retcode == 0
    print(output)
    assert verify_output(output, 
            "DEBUG: ### validated outputs", 
            "DEBUG: recipe 'alias test recipe'", 
            "DEBUG:   out: 1")

def test_test_nesting():
    print("===== expecting no errors =====")
    retcode, output = run("stimela -v -b native exec test_nesting.yml demo_recipe")
    assert retcode == 0
    print(output)


def test_test_recipe():
    print("===== expecting an error since 'msname' parameter is missing =====")
    retcode = os.system("stimela -v -b native exec test_recipe.yml selfcal.image_name=bar")
    assert retcode != 0 

    print("===== expecting an error due to elem-xyz choices wrong =====")
    retcode = os.system("stimela -v -b native exec test_recipe.yml selfcal.image_name=bar msname=foo elem-xyz=t elemlist-xyz=[x,y]")
    assert retcode != 0

    print("===== expecting an error due to elem-xyz choices wrong =====")
    retcode = os.system("stimela -v -b native exec test_recipe.yml selfcal.image_name=bar msname=foo elem-xyz=x elemlist-xyz=[x,t]")
    assert retcode != 0

    print("===== expecting no errors now =====")
    retcode = os.system("stimela -v -b native exec test_recipe.yml selfcal.image_name=bar msname=foo elem-xyz=x elemlist-xyz=[x,y]")
    assert retcode == 0

    print("===== expecting no errors now =====")
    retcode = os.system("stimela -v -b native exec test_recipe.yml selfcal.image_name=bar msname=foo elem-xyz=x elemlist-xyz=x")
    assert retcode == 0

def test_test_loop_recipe():
    print("===== expecting an error since 'ms' parameter is missing =====")
    retcode = os.system("stimela -v -b native exec test_loop_recipe.yml cubical_image_loop")
    assert retcode != 0

    print("===== expecting no errors now =====")
    retcode = os.system("stimela -v -b native exec test_loop_recipe.yml cubical_image_loop ms=foo")
    assert retcode == 0

    print("===== expecting no errors now =====")
    retcode = os.system("stimela -v -b native exec test_loop_recipe.yml same_as_cubical_image_loop ms=foo")
    assert retcode == 0

    for name in "abc":
        msname = f"test-{name}.ms"
        if not os.path.exists(msname):
            os.mkdir(msname)
    print("===== expecting no errors now =====")
    retcode = os.system("stimela -v -b native exec test_loop_recipe.yml loop_recipe")
    assert retcode == 0

def test_scatter():
    print("===== expecting no errors now =====")
    retcode = os.system("stimela -v -b native exec test_scatter.yml basic_loop")
    assert retcode == 0

    print("===== expecting no errors now =====")
    retcode = os.system("stimela -v -b native exec test_scatter.yml nested_loop")
    assert retcode == 0



import itertools
import click
import logging
import os.path
import yaml
import sys
import traceback
import re
import glob
import importlib
import itertools
from datetime import datetime
from typing import List, Optional, Tuple
from collections import OrderedDict
from omegaconf.omegaconf import OmegaConf, OmegaConfBaseException
from benedict import benedict


import stimela
from scabha import configuratt
from scabha.basetypes import UNSET
from scabha.exceptions import ScabhaBaseException
from scabha.substitutions import SubstitutionNS
from stimela import stimelogging
import stimela.config
from stimela.config import ConfigExceptionTypes
from stimela import logger, log_exception
from stimela.exceptions import RecipeValidationError, StimelaRuntimeError, StepSelectionError, StepValidationError
from stimela.main import cli
from stimela.kitchen.recipe import Recipe, Step, RecipeSchema, join_quote
from stimela import task_stats
import stimela.backends
from scabha.configuratt.common import IMPLICIT_EXTENSIONS

def resolve_recipe_files(filename: str, log: logging.Logger, use_manifest: bool = True):
    """
    Resolves a recipe file, which may be specified as (module)recipe.yml or module::recipe.yml, with
    the suffix being optional, and with wildcards allowed.

    If use_manifest is specified, looks for optional MANIFEST.stimela file.

    Returns list of real paths if file resolved, or None if filename should not be treated as a recipe file.

    Raises FileNotFoundError if filename refers to a recipe file that doesn't exist.
    """

    # check for (location)filename.yml or (location)/filename.yml style
    match1 = re.fullmatch(r"^\((.+)\)/?(.*)$", filename)
    match2 = re.fullmatch(r"^([\w.]+)::(.*)$", filename)
    if match1 or match2:
        modulename, fname = (match1 or match2).groups()
        try:
            mod = importlib.import_module(modulename)
        except ImportError as exc:
            raise FileNotFoundError(f"{filename} not found (can't import {modulename}: {exc})")
        # sort out module path
        if mod.__file__:
            resolved_filename = os.path.dirname(mod.__file__)
        else:
            try:
                resolved_filename = mod.__path__[0]
            except:
                raise FileNotFoundError(f"{filename} not found ({modulename} is not a proper Python module)")
        # append filename, if supplied
        if fname:
            resolved_filename = os.path.join(resolved_filename, fname)
            ext = os.path.splitext(fname)[1].lower()
        else:
            ext = None
    else:
        # else treat as filename, and check for extension
        ext = os.path.splitext(os.path.basename(filename))[1].lower()
        # no extension, and it doesn't look like a legit filename, return None
        if ext not in IMPLICIT_EXTENSIONS and "/" not in filename and filename not in (".", ".."):
            return None
        resolved_filename = filename

    # resolved to directory? Look for manifest or assume "*"
    if os.path.isdir(resolved_filename):
        manifest_file = os.path.join(resolved_filename, "MANIFEST.stimela")
        if use_manifest and os.path.isfile(manifest_file):
            globs = [os.path.join(resolved_filename, g.strip()) for g in open(manifest_file, "rt").readlines()]
            resolved_filename = manifest_file
            log.info(f"loading manifest from {manifest_file}")
        else:
            globs = [os.path.join(resolved_filename, "*" + e) for e in IMPLICIT_EXTENSIONS]
    else:
        # add extensions, if not supplied
        if not ext:
            # if name contains globs, search NAME.EXT, else search NAME and NAME.EXT
            globs = [] if resolved_filename.endswith("*") else [resolved_filename]
            globs += [resolved_filename + e for e in IMPLICIT_EXTENSIONS]
        else:
            globs = [resolved_filename]

    # apply globs and check for real files (i.e. don't match directories)
    filenames = itertools.chain(*[glob.glob(g) for g in globs])
    filenames = [f for f in filenames if os.path.isfile(f)]
        
    # return list or raise error
    if filenames:
        return filenames
    else:
        if filename is resolved_filename:
            raise FileNotFoundError(f"{filename} doesn't match any YAML documents")
        else:
            raise FileNotFoundError(f"{filename} resolves to {resolved_filename}, which doesn't match any YAML documents")


def load_recipe_files(filenames: List[str]):
    """Loads a set of config or recipe files. Returns list of recipes loaded."""

    full_conf = OmegaConf.create()
    full_deps = configuratt.ConfigDependencies()
    for filename in filenames:
        # check for (location)filename.yaml or (location)/filename.yaml style
        match1 = re.fullmatch(r"^\((.+)\)/?(.+)$", filename)
        match2 = re.fullmatch(r"^([\w.]+)::(.+)$", filename)
        if match1 or match2:
            modulename, filename = (match1 or match2).groups()
            try:
                mod = importlib.import_module(modulename)
            except ImportError as exc:
                log_exception(f"error importing {modulename}", exc)
                sys.exit(2)
            filename = os.path.join(os.path.dirname(mod.__file__), filename)
        # try loading
        try:
            conf, deps = configuratt.load(filename, use_sources=[stimela.CONFIG, full_conf], no_toplevel_cache=True)
        except FileNotFoundError as exc:
            log_exception(exc)
            sys.exit(2)
        except ConfigExceptionTypes as exc:
            log_exception(f"error loading {filename}", exc)
            sys.exit(2)
        # accumulate loaded config
        full_conf.merge_with(conf)
        full_deps.update(deps)

    # warn user if any includes failed
    failed_includes = {path: dep for path, dep in full_deps.fails.items() if dep.warn}
    if failed_includes:
        logger().warning(f"{len(failed_includes)} optional includes were not found")
        for path, dep in failed_includes.items():
            logger().warning(f"    {path} (from {dep.origin})")

    # merge into full config dependencies
    dependencies = stimela.config.get_initial_deps()
    dependencies.update(full_deps)
    stimela.config.CONFIG_DEPS.update(dependencies)

    # check for missing requirements
    missing = configuratt.check_requirements(full_conf, [stimela.CONFIG], strict=True)
    for (loc, name, _) in missing:
        logger().warning(f"optional config section '{loc}' omitted due to missing requirement '{name}'")

    # split content into config sections, and recipes:
    # config secions are merged into the config namespace, while recipes go under
    # lib.recipes
    recipe_names = []
    update_conf = OmegaConf.create()
    for name, value in full_conf.items():
        if name in stimela.CONFIG:
            update_conf[name] = value
        else:
            try:
                stimela.CONFIG.lib.recipes[name] = OmegaConf.merge(RecipeSchema, value)
            except Exception as exc:
                log_exception(f"error in definition of recipe '{name}'", exc)
                sys.exit(2)
            recipe_names.append(name)

    try:
        stimela.CONFIG.merge_with(update_conf)
    except Exception as exc:
        log_exception(f"error applying configuration from {' ,'.join(filenames)}", exc)
        sys.exit(2)

    return recipe_names

@cli.command("run",
    help="""
    Execute a single cab, or a recipe from a YML file.
    If the YML files contains multiple recipes, specify the recipe name as an extra argument.
    Use PARAM=VALUE to specify parameters for the recipe or cab. You can also use X.Y.Z=FOO to
    change any and all config and/or recipe settings.
    """,
    no_args_is_help=True)
@click.option("-s", "--step", "step_ranges", metavar="STEP(s)", multiple=True,
                help="""only runs specific step(s) from the recipe. Use commas, or give multiple times to cherry-pick steps.
                Use [BEGIN]:[END] to specify a range of steps. Note that cherry-picking an individual step via this option
                also impies --enable-step.""")
@click.option("-k", "--skip-step", "skip_ranges", metavar="STEP(s)", multiple=True,
                help="""forcefully skip specific recipe step(s). Use commas, or give multiple times to
                cherry-pick steps. Use [BEGIN]:[END] to specify a range of steps.""")
@click.option("-t", "--tags", "tags", metavar="TAG(s)", multiple=True,
                help="""only runs steps wth the given tags (and also steps tagged as "always").
                Use commas, or give multiple times for multiple tags.""")
@click.option("--skip-tags", "skip_tags", metavar="TAG(s)", multiple=True,
                help="""explicitly skips steps wth the given tags.
                Use commas, or give multiple times for multiple tags.""")
@click.option("-e", "--enable-step", "enable_steps", metavar="STEP(s)", multiple=True,
                help="""force-enable steps even if the recipe marks them as skipped. Use commas, or give multiple times
                for multiple steps.""")
@click.option("-f", "--disable-fresh-skips", "disable_fresh_skips",  is_flag=True,
                help="""forces execution of steps with a skip_if_outputs: fresh property.""")
@click.option("-F", "--disable-exist-skips", "disable_exist_skips",  is_flag=True,
                help="""forces execution of steps with a skip_if_outputs: exist property.""")
@click.option("-c", "--config", "config_equals", metavar="X.Y.Z=VALUE", nargs=1, multiple=True,
                help="""tweak configuration options.""")
@click.option("-a", "--assign", metavar="PARAM VALUE", nargs=2, multiple=True,
                help="""assigns values to parameters: equivalent to PARAM=VALUE, but plays nicer with the shell's
                tab completion feature.""")
@click.option("-C", "--config-assign", metavar="X.Y.Z VALUE", nargs=2, multiple=True,
                help="""tweak configuration options: same function -c/--config, but plays nicer with the shell's
                tab completion feature.""")
@click.option("-l", "--last-recipe", is_flag=True,
                help="""if multiple recipes are defined, selects the last one for execution.""")
@click.option("-d", "--dry-run", is_flag=True,
                help="""Doesn't actually run anything, only prints the selected steps.""")
@click.option("-p", "--profile", metavar="DEPTH", type=int,
                help="""Print per-step profiling stats to this depth. 0 disables.""")
@click.option("-N", "--native", "enable_native", is_flag=True,
                help="""Selects the native backend (shortcut for -C opts.backend.select=native)""")
@click.option("-S", "--singularity", "enable_singularity", is_flag=True,
                help="""Selects the singularity backend (shortcut for -C opts.backend.select=singularity)""")
@click.option("-K", "--kube", "enable_kube", is_flag=True,
                help="""Selects the kubernetes backend (shortcut for -C opts.backend.select=kube)""")
@click.option("--slurm", "enable_slurm", is_flag=True,
                help="""Enables the slurm backend wrapper (shortcut for -C backend.slurm.enable=True)""")
@click.option("-dc", "--dump-config", is_flag=True,
                help="""Dump the equivalent stimela config to a file""")
@click.option("-pf", "--parameter-file", metavar="filename.yml", multiple=True,
              help="""Use parameter values from the sepcified parameter file. These have lower precedence than
              PARAM=VALUE specified on the CLI.""")
@click.argument("parameters", nargs=-1, metavar="filename.yml ... [recipe or cab name] [PARAM=VALUE] ...", required=True)
def run(parameters: List[str] = [], dump_config: bool = False, dry_run: bool = False, last_recipe: bool = False, profile: Optional[int] = None,
    assign: List[Tuple[str, str]] = [],
    config_equals: List[str] = [],
    config_assign: List[Tuple[str, str]] = [],
    step_ranges: List[str] = [], tags: List[str] = [], skip_tags: List[str] = [], enable_steps: List[str] = [],
    skip_ranges: List[str] = [],
    disable_fresh_skips=False, disable_exist_skips=False,
    build=False, rebuild=False, build_skips=False,
    enable_native=False,
    enable_singularity=False,
    enable_kube=False,
    enable_slurm=False,
    parameter_file: List[str] = []
):

    log = logger()
    params = OrderedDict()
    errcode = 0
    recipe_or_cab = None
    files_to_load = []

    if disable_fresh_skips:
        stimela.CONFIG.opts.disable_skips.fresh = True
    if disable_exist_skips:
        stimela.CONFIG.opts.disable_skips.exist = True

    def convert_value(value):
        if value == "=UNSET":
            return UNSET
        else:
            return yaml.safe_load(value)

    # Load parameter files using benedict and flatten into a dotstring-value dict.
    for pf in parameter_file:
        pf_dict = benedict.from_yaml(pf, keypath_separator=">>").flatten(separator=".")
        for key, value in pf_dict.items():
            try:
                params[key] = value
            except Exception as exc:
                log_exception(f"error parsing {key} {value} in parameter file {pf}", exc)
                errcode = 2

    # parse assign values as YaML
    for key, value in assign:
        # parse string as yaml value
        try:
            params[key] = convert_value(value)
        except Exception as exc:
            log_exception(f"error parsing value for --assign {key} {value}", exc)
            errcode = 2

    # parse arguments as recipe name, parameter assignments, or dotlist for OmegaConf
    for pp in parameters:
        if "=" in pp:
            key, value = pp.split("=", 1)
            # parse string as yaml value
            try:
                params[key] = convert_value(value)
            except Exception as exc:
                log_exception(f"error parsing {pp}", exc)
                errcode = 2
        else:
            try:
                filenames = resolve_recipe_files(pp, log=log)
            except FileNotFoundError as exc:
                log_exception(exc)
                errcode = 2
                continue
            if filenames is not None:
                files_to_load += filenames
                log.info(f"will load recipe/config file(s) {', '.join(filenames)}")
            elif recipe_or_cab is not None:
                log_exception(f"multiple recipe/cab names given")
                errcode = 2
            else:
                recipe_or_cab = pp
                log.info(f"treating '{pp}' as a recipe or cab name")

    if errcode:
        sys.exit(errcode)

    # load config and recipes from all given files
    if files_to_load:
        available_recipes = load_recipe_files(files_to_load)
    else:
        available_recipes = []

    # load config settigs from --config arguments
    try:
        stimela.CONFIG.merge_with(OmegaConf.from_dotlist(config_equals))
    except OmegaConfBaseException as exc:
        log_exception(f"error loading -c/--config assignments", exc)
        sys.exit(2)
    try:
        dotlist = [f"{key}={value}" for key, value in config_assign]
        stimela.CONFIG.merge_with(OmegaConf.from_dotlist(dotlist))
    except OmegaConfBaseException as exc:
        log_exception(f"error loading -C/--config-assign assignments", exc)
        sys.exit(2)

    # enable backends
    if enable_native:
        log.info("selecting the native backend")
        stimela.CONFIG.opts.backend.select = 'native'
    elif enable_singularity:
        log.info("selecting the singularity backend")
        stimela.CONFIG.opts.backend.select = 'singularity'
    elif enable_kube:
        log.info("selecting the kube backend")
        stimela.CONFIG.opts.backend.select = 'kube'
    if enable_slurm:
        log.info("enabling the slurm backend wrapper")
        stimela.CONFIG.opts.backend.slurm.enable = True

    def log_available_runnables():
        """Helper function to list available recipes or cabs"""
        if available_recipes:
            log.info(f"available recipes: {' '.join(available_recipes)}")
        if stimela.CONFIG.cabs:
            log.info(f"available cabs: {' '.join(stimela.CONFIG.cabs.keys())}")

    # figure out what we're running, recipe or cab
    recipe_name = cab_name = None
    # do we need to make an implicit choice?
    if recipe_or_cab is None:
        # -l specified, pick the last recipe
        if last_recipe:
            if not available_recipes:
                log.error(f"-l/--last-recipe specified, but no valid recipes were loaded")
                sys.exit(2)
            else:
                recipe_name = available_recipes[-1]
                log.info(f"-l/--last-recipe specified, selecting '{recipe_name}'")
        # nothing specified, either we have exactly 1 recipe defined (pick that), or 0 recipes and 1 cab
        elif len(available_recipes) == 1:
            recipe_name = available_recipes[0]
            log.info(f"found single recipe '{recipe_name}', selecting it implicitly")
        elif len(stimela.CONFIG.cabs) == 1 and not available_recipes:
            cab_name = next(iter(stimela.CONFIG.cabs))
            log.info(f"found single cab '{cab_name}', selecting it implicitly")
        else:
            log.error("found multiple recipes or cabs, please specify one on the command line")
            log_available_runnables()
            sys.exit(2)
    # else something was specified
    elif recipe_or_cab in available_recipes:
        recipe_name = recipe_or_cab
        log.info(f"selected recipe is '{recipe_name}'")
    elif recipe_or_cab in stimela.CONFIG.cabs:
        cab_name = recipe_or_cab
        log.info(f"selected cab is '{cab_name}'")
    else:
        if not available_recipes and not stimela.CONFIG.cabs:
            log.error("no valid recipes or cabs were loaded")
        else:
            log.error(f"'{recipe_or_cab}' does not refer to a recipe or a cab")
            log_available_runnables()
        sys.exit(2)

    # are we running a standalone cab?
    if cab_name is not None:
        # create step config by merging in settings (var=value pairs from the command line)
        outer_step = Step(cab=cab_name, params=params)
        outer_step.name = cab_name
        # provide basic substitutions for running the step below
        subst = SubstitutionNS()
        info = SubstitutionNS(fqname=cab_name, label=cab_name, label_parts=[], suffix='', taskname=cab_name)
        subst._add_('info', info, nosubst=True)
        subst._add_('config', stimela.CONFIG, nosubst=True)
        subst._add_('current', SubstitutionNS(**params))
        # create step logger manually, since we won't be doing the normal recipe-level log management
        step_logger = stimela.logger().getChild(cab_name)
        step_logger.propagate = True
        try:
            outer_step.finalize(fqname=cab_name, log=step_logger)
            outer_step.prevalidate(root=True, subst=subst)
        except ScabhaBaseException as exc:
            log_exception(exc)
            sys.exit(1)
        # check for missing and unresolved parameters
        if not build and (outer_step.missing_params or outer_step.unresolved_params):
            missing = {}
            for name in outer_step.missing_params:
                schema = outer_step.inputs_outputs[name]
                if schema.is_input or schema.is_named_output:
                    missing[name] = schema.info
            # don't report unresolved implicits, since that's just a consequence of a missing input
            for name in outer_step.unresolved_params:
                schema = outer_step.inputs_outputs[name]
                if (schema.is_input or schema.is_named_output) and not schema.implicit:
                    missing[name] = outer_step.inputs_outputs[name].info
            # print missing parameters and exit
            if missing:
                log_exception(StepValidationError(f"cab '{cab_name}' is missing required parameter(s)", missing))
                sys.exit(1)

    # else run a recipe
    else:
        # create recipe object from the config
        kwargs = dict(**stimela.CONFIG.lib.recipes[recipe_name])
        kwargs.setdefault('name', recipe_name)
        try:
            recipe = Recipe(**kwargs)
        except Exception as exc:
            traceback.print_exc()
            log_exception(f"error loading recipe '{recipe_name}'", exc)
            sys.exit(2)

        # force name, if not set
        if not recipe.name:
            recipe.name = recipe_name

        # wrap it in an outer step and prevalidate (to set up loggers etc.)
        recipe.fqname = recipe_name
        try:
            recipe.finalize()
        except Exception as exc:
            log_exception(RecipeValidationError(f"error validating recipe '{recipe_name}'", exc))
            for line in traceback.format_exc().split("\n"):
                log.debug(line)
            sys.exit(1)

        for key, value in params.items():
            try:
                recipe.assign_value(key, value, override=True)
            except ScabhaBaseException as exc:
                log_exception(exc)
                sys.exit(1)

        # split out parameters
        params = {key: value for key, value in params.items() if key in recipe.inputs_outputs}

        stimelogging.declare_chapter("prevalidation")
        log.info("pre-validating the recipe")
        outer_step = Step(recipe=recipe, name=f"{recipe_name}", info=recipe_name, params=params)
        try:
            params = outer_step.prevalidate(root=True)
        except Exception as exc:
            log_exception(RecipeValidationError(f"pre-validation of recipe '{recipe_name}' failed", exc))
            for line in traceback.format_exc().split("\n"):
                log.debug(line)
            sys.exit(1)

        # select recipe substeps based on command line, and exit if nothing to run
        if not build_skips:
            selection_options = []
            for opts in (tags, skip_tags, step_ranges, skip_ranges, enable_steps):
                selection_options.append(set(itertools.chain(*(opt.split(",") for opt in opts))))

            try:
                if not recipe.restrict_steps(*selection_options):
                    sys.exit(0)
            except StepSelectionError as exc:
                log_exception(exc)
                sys.exit(2)

        logdir = stimelogging.get_logfile_dir(recipe.log) or '.'
        log.info(f"recipe logs will be saved under {logdir}")

        filename = os.path.join(logdir, "stimela.recipe.deps")
        stimela.config.CONFIG_DEPS.save(filename)
        log.info(f"saved recipe dependencies to {filename}")

        # no substitutions provided, recipe initializes its own
        subst = None

    # in debug mode, pretty-print the recipe
    if log.isEnabledFor(logging.DEBUG):
        log.debug("---------- prevalidated step follows ----------")
        for line in outer_step.summary(params=params):
            log.debug(line)

    if dump_config:
        filename = os.path.join(logdir, "stimela.recipe.config.yaml")
        log.info(f"recipe config will be saved under {filename}")
        OmegaConf.save(stimela.CONFIG, f=filename)

    if dry_run:
        log.info("dry run was requested, exiting")
        sys.exit(0)

    start_time = datetime.now()
    def elapsed():
        return str(datetime.now() - start_time).split('.', 1)[0]

    # build the images
    if build:
        try:
            outer_step.build(rebuild=rebuild, build_skips=build_skips, log=log)
        except Exception as exc:
            stimela.backends.close_backends(log)

            if not isinstance(exc, ScabhaBaseException) or not exc.logged:
                log_exception(StimelaRuntimeError(f"build failed after {elapsed()}", exc,
                    tb=not isinstance(exc, ScabhaBaseException)))
            else:
                log.error("build failed, exiting with error code 1")
            for line in traceback.format_exc().split("\n"):
                log.debug(line)
            last_log_dir = stimelogging.get_logfile_dir(outer_step.log) or '.'
            outer_step.log.info(f"last log directory was {stimelogging.apply_style(last_log_dir, 'bold green')}")
            sys.exit(1)

    # else run the recipe
    else:
        try:
            outputs = outer_step.run(is_outer_step=True, subst=subst)
        except Exception as exc:
            stimela.backends.close_backends(log)

            task_stats.save_profiling_stats(outer_step.log,
                print_depth=profile if profile is not None else stimela.CONFIG.opts.profile.print_depth,
                unroll_loops=stimela.CONFIG.opts.profile.unroll_loops)
            if not isinstance(exc, ScabhaBaseException) or not exc.logged:
                log_exception(StimelaRuntimeError(f"run failed after {elapsed()}", exc,
                    tb=not isinstance(exc, ScabhaBaseException)))
            else:
                log.error("run failed, exiting with error code 1")
            for line in traceback.format_exc().split("\n"):
                log.debug(line)
            last_log_dir = stimelogging.get_logfile_dir(outer_step.log) or '.'
            outer_step.log.info(f"last log directory was {stimelogging.apply_style(last_log_dir, 'bold green')}")
            sys.exit(1)

        if outputs and outer_step.log.isEnabledFor(logging.DEBUG):
            outer_step.log.debug(f"run successful after {elapsed()}, outputs follow:")
            for name, value in outputs.items():
                if name in outer_step.outputs:
                    outer_step.log.debug(f"  {name}: {value}")
        else:
            outer_step.log.info(f"run successful after {elapsed()}")

    stimela.backends.close_backends(log)

    if not build:
        task_stats.save_profiling_stats(outer_step.log,
                print_depth=profile if profile is not None else stimela.CONFIG.opts.profile.print_depth,
                unroll_loops=stimela.CONFIG.opts.profile.unroll_loops)

    last_log_dir = stimelogging.get_logfile_dir(outer_step.log) or '.'
    outer_step.log.info(f"last log directory was {stimelogging.apply_style(last_log_dir, 'bold green')}")
    return 0



import select 
import traceback
import subprocess
import errno
import re
import time
import logging
import sys
import signal

DEBUG = 0
from stimela.exceptions import StimelaCabRuntimeError, StimelaProcessRuntimeError

log = None

from .xrun_asyncio import get_stimela_logger, dispatch_to_log

def global_logger():
    """Returns Stimela logger if running in stimela, else inits a global logger"""
    global log
    if log is None:
        log = get_stimela_logger()
        if log is None:
            # no stimela => running payload inside a cab -- just use the global logger and make it echo everything to the console
            logging.basicConfig(format="%(message)s", level=logging.INFO, stream=sys.stdout)
            log = logging.getLogger()
    return log

def xrun_nolog(command, name=None, shell=True):
    log = global_logger()
    name = name or command.split(" ", 1)[0]
    try:
        log.info("# running {}".format(command))
        status = subprocess.call(command, shell=shell)

    except KeyboardInterrupt:
        log.error("# {} interrupted by Ctrl+C".format(name))
        raise

    except Exception as exc:
        for line in traceback.format_exc():
            log.error("# {}".format(line.strip()))
        log.error("# {} raised exception: {}".format(name, str(exc)))
        raise

    if status:
        raise StimelaProcessRuntimeError("{} returns error code {}".format(name, status))

    return 0


class SelectPoller(object):
    """Poller class. Poor man's select.poll(). Damn you OS/X and your select.poll will-you-won'y-you bollocks"""
    def __init__ (self, log):
        self.fdlabels = {}
        self.log = log

    def register_file(self, fobj, label):
        self.fdlabels[fobj.fileno()] = label, fobj

    def register_process(self, po, label_stdout='stdout', label_stderr='stderr'):
        self.fdlabels[po.stdout.fileno()] = label_stdout, po.stdout
        self.fdlabels[po.stderr.fileno()] = label_stderr, po.stderr

    def poll(self, timeout=5, verbose=False):
        while True:
            try:
                to_read, _, _ = select.select(self.fdlabels.keys(), [], [], timeout)
                self.log.debug("poll(): ready to read: {}".format(to_read))
                # return on success or timeout
                return [self.fdlabels[fd] for fd in to_read]
            except (select.error, IOError) as ioerr:
                if verbose:
                    self.log.debug("poll() exception: {}".format(traceback.format_exc()))
                if hasattr(ioerr, 'args'):
                    err = ioerr.args[0]  # py2
                else:
                    err = ioerr.errno    # py3
                # catch interrupted system call -- return if we have a timeout, else
                # loop again
                if err == errno.EINTR:
                    if timeout is not None:
                        if verbose:
                            self.log.debug("poll(): returning")
                        return []
                    if verbose:
                        self.log.debug("poll(): retrying")
                else:
                    raise ioerr

    def unregister_file(self, fobj):
        if fobj.fileno() in self.fdlabels:
            del self.fdlabels[fobj.fileno()]

    def __contains__(self, fobj):
        return fobj.fileno() in self.fdlabels

class Poller(object):
    """Poller class. Wraps select.poll()."""
    def __init__ (self, log):
        self.fdlabels = {}
        self.log = log
        self._poll = select.poll()

    def register_file(self, fobj, label):
        self.fdlabels[fobj.fileno()] = label, fobj
        self._poll.register(fobj.fileno(), select.POLLIN)

    def register_process(self, po, label_stdout='stdout', label_stderr='stderr'):
        self.fdlabels[po.stdout.fileno()] = label_stdout, po.stdout
        self.fdlabels[po.stderr.fileno()] = label_stderr, po.stderr
        self._poll.register(po.stdout.fileno(), select.POLLIN)
        self._poll.register(po.stderr.fileno(), select.POLLIN)

    def poll(self, timeout=5, verbose=False):
        try:
            to_read = self._poll.poll(timeout*1000)
            if verbose:
                self.log.debug("poll(): ready to read: {}".format(to_read))
            return [self.fdlabels[fd] for (fd, ev) in to_read]
        except Exception:
            if verbose:
                self.log.debug("poll() exception: {}".format(traceback.format_exc()))
            raise

    def unregister_file(self, fobj):
        if fobj.fileno() in self.fdlabels:
            self._poll.unregister(fobj.fileno())
            del self.fdlabels[fobj.fileno()]

    def __contains__(self, fobj):
        return fobj.fileno() in self.fdlabels



def xrun(command, options, log=None, env=None, timeout=-1, kill_callback=None, output_wrangler=None, shell=True, return_errcode=False, command_name=None, log_command=True):
    command_name = command_name or command

    # this part could be inside the container
    command_line = " ".join([command] + list(map(str, options)))
    if shell:
        command_line = " ".join([command] + list(map(str, options)))
        command = [command_line]
    else:
        command = [command] + list(map(str, options))
        command_line = " ".join(command)

    log = log or get_stimela_logger()

    if log is None:
        return xrun_nolog(command, name=command_name, shell=shell)

    # this part is never inside the container
    import stimela

    log = log or stimela.logger()

    if log_command:
        log.info("running " + command_line, extra=dict(stimela_subprocess_output=(command_name, "start")))

    start_time = time.time()

    proc = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                            env=env, bufsize=1, universal_newlines=True, shell=shell)

    poller = Poller(log=log)
    poller.register_process(proc)

    proc_running = True

    try:
        while proc_running and poller.fdlabels:
            fdlist = poller.poll(verbose=DEBUG>0)
#            print(f"fdlist is {fdlist}")
            for fname, fobj in fdlist:
                try:
                    line = fobj.readline()
                except EOFError:
                    line = b''
#                print("read {} from {}".format(line, fname))
                empty_line = not line
                line = (line.decode('utf-8') if type(line) is bytes else line).rstrip()
                # break out if process closes
                if empty_line:
                    poller.unregister_file(fobj)
                    if proc.stdout not in poller and proc.stderr not in poller:
                        log.debug(f"the {command_name} process has exited")
                        proc_running = None
                        break
                    continue
                # dispatch output to log
                dispatch_to_log(log, line, command_name, 
                                stream_name="stderr" if fobj is proc.stderr else "stdout", 
                                output_wrangler=output_wrangler)
            if timeout > 0 and time.time() > start_time + timeout:
                log.error(f"timeout, killing {command_name} process")
                kill_callback() if callable(kill_callback) else proc.kill()
                proc_running = False

        proc.wait()
        status = proc.returncode

    except SystemExit as exc:
        proc.wait()
        status = exc.code
        raise StimelaCabRuntimeError(f"{command_name}: SystemExit with code {status}", log=log)

    except KeyboardInterrupt:
        if callable(kill_callback):
            log.warning(f"Ctrl+C caught: shutting down {command_name} process, please give it a few moments")
            kill_callback() 
            log.info(f"the {command_name} process was shut down successfully",
                     extra=dict(stimela_subprocess_output=(command_name, "status")))
            proc.wait()
        else:
            log.warning(f"Ctrl+C caught, interrupting {command_name} process {proc.pid}")
            proc.send_signal(signal.SIGINT)
            for retry in range(10):
                if retry:
                    log.info(f"Process {proc.pid} not exited after {retry} seconds, waiting a bit longer...")
                try:
                    proc.wait(1)
                    log.info(f"Process {proc.pid} has exited with return code {proc.returncode}")
                    break
                except subprocess.TimeoutExpired as exc:
                    if retry == 4:
                        log.warning(f"Terminating process {proc.pid}")
                        proc.terminate()
            else:
                log.warning(f"Killing process {proc.pid}")
                proc.kill()
                proc.wait()
    
        raise StimelaCabRuntimeError(f"{command_name} interrupted with Ctrl+C")

    except Exception as exc:
        traceback.print_exc()
        proc.wait()
        raise StimelaCabRuntimeError(f"{command_name} threw exception: {exc}'", log=log)

    if status and not return_errcode:
        raise StimelaCabRuntimeError(f"{command_name} returns error code {status}")
    
    return status
    



#!/usr/bin/env python
from typing import Tuple, List, Optional
import sys
import os.path
import click
from scabha.schema_utils import clickify_parameters

schema_file = os.path.join(os.path.dirname(__file__), "test_clickify.yaml")

@click.command()
@clickify_parameters(schema_file)
def func(name: str, i: int, j: Optional[float] = 1,
         remainder: Optional[List[str]] = None, 
         k: float=2,
         tup: Optional[Tuple[int, str]] = None, 
         files1: Optional[List[str]] = None,
         files2: Optional[List[str]] = None,
         files3: Optional[List[str]] = None,
         output: str = None):
    print(f"name:{name} i:{i} j:{j} k:{k} tup:{tup}")
    print(f"remainder: {remainder}")
    print(f"files1: {files1}")
    print(f"files2: {files2}")
    print(f"files3: {files3}")
    print(f"output: {output}")

if __name__ == "__main__":
    sys.exit(func())





import glob
import os.path
import fnmatch
import pyparsing
pyparsing.ParserElement.enable_packrat()
from pyparsing import *
from pyparsing import common
from functools import reduce
import operator
import dataclasses

from .substitutions import SubstitutionError, SubstitutionContext
from .basetypes import Unresolved, UNSET
from .exceptions import *

import typing
from typing import Dict, List, Any
from omegaconf import DictConfig, ListConfig

_parser = None

# see https://stackoverflow.com/questions/43244861/pyparsing-infixnotation-optimization for a cleaner parser with functions

def _not_operator(value):
    return value is UNSET or isinstance(value, Unresolved) or not value

_UNARY_OPERATORS = {
    '+':    lambda x: +x,
    '-':    lambda x: -x,
    '~':    lambda x: ~x,
    'not':  _not_operator
}

_BINARY_OPERATORS = {
    '**':   lambda x,y: x ** y,
    '*':   lambda x,y: x * y,
    '/':   lambda x,y: x / y,
    '//':   lambda x,y: x // y,
    '+':   lambda x,y: x + y,
    '-':   lambda x,y: x - y,
    '<<':   lambda x,y: x << y,
    '>>':   lambda x,y: x >> y,
    '&':   lambda x,y: x & y,
    '^':   lambda x,y: x ^ y,
    '|':   lambda x,y: x | y,
    '==':   lambda x,y: x == y,
    '!=':   lambda x,y: x != y,
    '<=':   lambda x,y: x <= y,
    '<':   lambda x,y: x < y,
    '>=':   lambda x,y: x >= y,
    '>':   lambda x,y: x > y,
    # 'is':   lambda x,y: x is y,
    # 'is not':   lambda x,y: x is not y,
    'in':   lambda x,y: x in y,
    'not in':   lambda x,y: x not in y,
    'and':   lambda x,y: x and y,
    'or':   lambda x,y: x or y,
}        


class ResultsHandler(object):
    def evaluate(evaluator):
        pass

    def dump(self):
        pass

    @classmethod
    def pa(cls, s, l, t):
        return cls(*t[0])


class UnaryHandler(ResultsHandler):
    def __init__(self, op, arg):
        self.op, self.arg = op, arg
        assert op in _UNARY_OPERATORS
        self._op = _UNARY_OPERATORS[op]
        # not allows UNSET values
        self.allow_unset = (op == 'not')
    
    def evaluate(self, evaluator):
        arg = evaluator._evaluate_result(self.arg, allow_unset=self.allow_unset)
        if isinstance(arg, Unresolved):
            return arg
        return self._op(arg)

    def dump(self):
        return f"UnaryHandler({self.op}\n   {self.arg})"


class BinaryHandler(ResultsHandler):
    def __init__(self, arg1, op, arg2):
        self.op, self.arg1, self.arg2 = op, arg1, arg2
        assert op in _BINARY_OPERATORS
        self._op = _BINARY_OPERATORS[op]
        self.allow_unset = False
    
    def evaluate(self, evaluator):
        arg1, arg2 = evaluator._evaluate_result(self.arg1, allow_unset=self.allow_unset), \
                     evaluator._evaluate_result(self.arg2, allow_unset=self.allow_unset)
        if isinstance(arg1, Unresolved):
            return arg1
        if isinstance(arg2, Unresolved):
            return arg2
        return self._op(arg1, arg2)

    def dump(self):
        return f"BinaryHandler({self.arg1},\n   {self.op},\n{self.arg2})"

    @staticmethod
    def pa(s, l, t):
        # https://stackoverflow.com/questions/4571441/recursive-expressions-with-pyparsing
        initlen, incr = 3, 2
        t = t[0]
        ret = BinaryHandler(*t[:initlen])
        i = initlen
        while i < len(t):
            ret = BinaryHandler(ret, *t[i:i+incr])
            i += incr
        return ret

class GetItemHandler(ResultsHandler):
    def __init__(self, base):
        self.base, self.index = base[0], base[1]

    @staticmethod
    def pa(s, l, t):
        # https://stackoverflow.com/questions/4571441/recursive-expressions-with-pyparsing
        return GetItemHandler(*t)

    def evaluate(self, evaluator):
        base = evaluator._evaluate_result(self.base)
        index = evaluator._evaluate_result(self.index)
        if isinstance(base, Unresolved):
            return base
        if isinstance(index, Unresolved):
            return index
        return base[index]

class FunctionHandler(ResultsHandler):
    def __init__(self, func, *args):
        self.func, self.args = func, args
        self._func = getattr(self, func)
        if self._func is None:
            raise FormulaError(f"unknown function {func}")

    @staticmethod
    def pa(s, l, t):
        # https://stackoverflow.com/questions/4571441/recursive-expressions-with-pyparsing
        return FunctionHandler(*t[0])

    def evaluate(self, evaluator):
        return self._func(evaluator, self.args)

    def evaluate_generic_callable(self, evaluator, name, callable, args, min_args=None, max_args=None):
        if min_args is not None and len(args) < min_args:
            raise FormulaError(f"{'.'.join(evaluator.location)}: {name}() expects at least {min_args} argument(s)")
        if max_args is not None and len(args) > max_args:
            raise FormulaError(f"{'.'.join(evaluator.location)}: {name}() expects at most {max_args} argument(s)")
        eval_args = [evaluator._evaluate_result(arg) for arg in args]
        # if any argument is UNSET, return it as our result
        unsets = [arg for arg in eval_args if isinstance(arg, Unresolved)]
        if unsets:
            return unsets[0]
        return callable(*eval_args)

    def ERROR(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: ERROR() expects one argument, got {len(args)}")
        cond = evaluator._evaluate_result(args[0], allow_unset=True)
        raise FormulaError(f"ERROR: {cond}")

    def LIST(self, evaluator, args):
        def make_list(*x):
            return list(x)
        return self.evaluate_generic_callable(evaluator, "LIST", make_list, args)

    def LEN(self, evaluator, args):
        def make_len(x):
            return len(x)
        return self.evaluate_generic_callable(evaluator, "LEN", make_len,
                                              args, min_args=1, max_args=1)

    def RANGE(self, evaluator, args):
        def make_range(*x):
            return list(range(*x))
        return self.evaluate_generic_callable(evaluator, "RANGE", make_range, args, min_args=1, max_args=3)
    
    def VALID(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: VALID() expects one argument, got {len(args)}")
        try:
            result = evaluator._evaluate_result(args[0], allow_unset=True)
        except (TypeError, ValueError) as exc:
            return False
        if isinstance(result, Unresolved):
            return False        
        return result

    def TRY(self, evaluator, args):
        for arg in args:
            result = evaluator._evaluate_result(arg, allow_unset=True)
            if not isinstance(result, Unresolved):
                return result
        return result

    def MIN(self, evaluator, args):
        return self.evaluate_generic_callable(evaluator, "MIN", min, args, min_args=1)

    def MAX(self, evaluator, args):
        return self.evaluate_generic_callable(evaluator, "MAX", max, args, min_args=1)
    
    def GETITEM(self, evaluator, args):
        def get_item(x, y):
            return x[y]
        return self.evaluate_generic_callable(evaluator, "GETITEM", get_item, args, min_args=2, max_args=2)

    def IS_STR(self, evaluator, args):
        def is_str(x):
            return type(x) is str
        return self.evaluate_generic_callable(evaluator, "IS_STR", is_str, args, min_args=1, max_args=1)
    
    def IS_NUM(self, evaluator, args):
        def is_num(x):
            return isinstance(x, (bool, int, float, complex)) 
        return self.evaluate_generic_callable(evaluator, "IS_NUM", is_num, args, min_args=1, max_args=1)
    
    def CASES(self, evaluator, args):
        # set default case
        if len(args)%2:
            default_case = args[-1]
            args = args[:-1]
        else: 
            default_case = None
        # return first True case
        for i in range(0, len(args), 2):
            conditional, result = args[i:i+2]
            cond = evaluator._evaluate_result(conditional, allow_unset=False)
            if cond:
                return evaluator._evaluate_result(result, allow_unset=True)
        # return default
        if default_case is None:
            return UNSET("no match in CASES()")
        return evaluator._evaluate_result(default_case, allow_unset=True)

    def IF(self, evaluator, args):
        if len(args) < 3 or len(args) > 4:
            raise FormulaError(f"{'.'.join(evaluator.location)}: IF() expects 3 or 4 arguments, got {len(args)}")
        conditional, if_true, if_false = args[:3]
        if_unset = args[3] if len(args) == 4 else UNSET("")

        cond = evaluator._evaluate_result(conditional, allow_unset=if_unset is not None)
        if isinstance(cond, Unresolved):
            if if_unset is None:
                raise SubstitutionError(f"{'.'.join(evaluator.location)}: '{cond.name}' is not defined")
            result = if_unset
        else:
            result = if_true if cond else if_false    
            
        return evaluator._evaluate_result(result)

    def IFSET(self, evaluator, args):
        if len(args) < 1 or len(args) > 3:
            raise FormulaError(f"{'.'.join(evaluator.location)}: IFSET() expects 1 to 3 arguments, got {len(args)}")
        
        lookup, if_set, if_unset = list(args) + [None]*(3 - len(args))

        value = evaluator._evaluate_result(lookup, allow_unset=True)
        if isinstance(value, Unresolved):
            if is_missing(if_unset):
                return UNSET
            else:
                return evaluator._evaluate_result(if_unset)
        elif is_missing(if_set) or if_set == 'SELF':
            return value
        else:
            return evaluator._evaluate_result(if_set)            

    def GLOB(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: GLOB() expects 1 argument, got {len(args)}")
        pattern = evaluator._evaluate_result(args[0])
        if isinstance(pattern, Unresolved):
            return pattern
        return sorted(glob.glob(pattern))

    def EXISTS(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: EXISTS() expects 1 argument, got {len(args)}")
        pattern = evaluator._evaluate_result(args[0])
        if isinstance(pattern, Unresolved):
            return pattern
        return bool(glob.glob(pattern))

    def DIRNAME(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: DIRNAME() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if isinstance(path, Unresolved):
            return path
        if not isinstance(path, str):
            raise FormulaError(f"DIRNAME() expects a string, got a {str(type(path))}") 
        return os.path.dirname(path)

    def BASENAME(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: BASENAME() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if isinstance(path, Unresolved):
            return path
        if not isinstance(path, str):
            raise FormulaError(f"BASENAME() expects a string, got a {str(type(path))}") 
        return os.path.basename(path)

    def EXTENSION(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: EXTENSION() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if isinstance(path, Unresolved):
            return path
        if not isinstance(path, str):
            raise FormulaError(f"EXTENSION() expects a string, got a {str(type(path))}") 
        return os.path.splitext(path)[1]

    def STRIPEXT(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: STRIPEXT() expects 1 argument, got {len(args)}")
        path = evaluator._evaluate_result(args[0])
        if isinstance(path, Unresolved):
            return path
        if not isinstance(path, str):
            raise FormulaError(f"STRIPEXT() expects a string, got a {str(type(path))}") 
        return os.path.splitext(path)[0]

    def NOSUBST(self, evaluator, args):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: NOSUBST() expects 1 argument, got {len(args)}")
        return evaluator._evaluate_result(args[0], subst=False)
    
    def SORT(self, evaluator, args):
        return self._sort_impl(evaluator, args, "SORT")
    
    def RSORT(self, evaluator, args):
        return self._sort_impl(evaluator, args, "RSORT", reverse=True)

    def _sort_impl(self, evaluator, args, funcname, reverse=False):
        if len(args) != 1:
            raise FormulaError(f"{'.'.join(evaluator.location)}: {funcname}() expects 1 argument, got {len(args)}")
        sortlist = evaluator._evaluate_result(args[0])
        if isinstance(sortlist, Unresolved):
            return sortlist
        if not isinstance(sortlist, (list, tuple)):
            raise FormulaError(f"{funcname}() expects a list, got a {str(type(sortlist))}") 
        return sorted(sortlist, reverse=reverse)
    

def construct_parser():
    lparen = Literal("(").suppress()
    rparen = Literal(")").suppress()
    lbrack = Literal("[").suppress()
    rbrack = Literal("]").suppress()
    comma = Literal(",").suppress()
    period = Literal(".").suppress()
    string = (QuotedString('"') | QuotedString("'"))("constant")
    UNSET = Keyword("UNSET")("unset")
    SELF = Keyword("SELF")("self_value")
    EMPTY = Keyword("EMPTY")("empty")
    bool_false = (Keyword("False") | Keyword("false"))("bool_false").set_parse_action(lambda:[False])
    bool_true = (Keyword("True") | Keyword("true"))("bool_true").set_parse_action(lambda:[True])

    boolean = (bool_true | bool_false)("constant")
    number = common.number("constant")

    fieldname = Word(alphas + "_", alphanums + "_-@*?")
    nested_field = Group(fieldname + OneOrMore(period + fieldname))("namespace_lookup")
    anyseq = CharsNotIn(",)")("constant")

    # allow expression to be used recursively
    expr = Forward()
    
    # functions -- get all all-uppercase members from FunctionHandler
    anyseq_funcnames = {"GLOB", "EXISTS", "ERROR"}
    all_funcnames = set(func for func in dir(FunctionHandler) 
                        if callable(getattr(FunctionHandler, func)) and func.upper() == func)
    all_funcnames -= anyseq_funcnames

    functions = reduce(operator.or_, map(Keyword, all_funcnames))
    # these functions take one argument, which could also be a sequence
    anyseq_functions = reduce(operator.or_, map(Keyword, anyseq_funcnames))

    atomic_value = (boolean | UNSET | EMPTY | nested_field | string | number)

    function_call_anyseq = Group(anyseq_functions + lparen + (expr | anyseq) + rparen).setParseAction(FunctionHandler.pa)
    function_call = Group(functions + lparen + 
                    Opt(delimited_list(expr|SELF)) + 
                    rparen).setParseAction(FunctionHandler.pa)

    operators = [
        ((lbrack + expr + rbrack), 1, opAssoc.LEFT, GetItemHandler.pa),
        (Literal("**"), 2, opAssoc.LEFT, BinaryHandler.pa), 
        (Literal("-")|Literal("+")|Literal("~"), 1, opAssoc.RIGHT, UnaryHandler.pa), 
        (Literal("*")|Literal("//")|Literal("/")|Literal("%"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("+")|Literal("-"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("<<")|Literal(">>"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("&"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("^"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (Literal("|"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (reduce(operator.or_, map(Literal, ("==", "!=", ">=", "<=", ">", "<"))), 2, opAssoc.LEFT, BinaryHandler.pa),
        (CaselessKeyword("in")|CaselessKeyword("not in"), 2, opAssoc.LEFT, BinaryHandler.pa),
        (CaselessKeyword("not"), 1, opAssoc.RIGHT, UnaryHandler.pa),
        (CaselessKeyword("and")|CaselessKeyword("or"), 2, opAssoc.LEFT, BinaryHandler.pa),
    ]

    infix = infix_notation(atomic_value | function_call | function_call_anyseq | nested_field,
                            operators)("subexpression")

    expr <<= infix

    return expr


if _parser is None:
    _parser = construct_parser()

_parse_cache = {}

def parse_string(text: str, location: List[str] = []):
    """parses a formula, returns ParseResults object. Implements cache.

    Args:
        text (str): formula to parse
        location (List[str], optional): Hierarchical location, used for erro messages
        
    Raises:
        ParserError: on any parse error

    Returns:
        ParseResults: on successful parse
    """
    parse_results = _parse_cache.get(text)
    if parse_results is None:
        try:
            parse_results = _parser.parse_string(text, parse_all=True)
        except Exception as exc:
            parse_results = ParserError(f"{'.'.join(location)}: error parsing formula ({exc})")
        _parse_cache[text] = parse_results

    if isinstance(parse_results, Exception):
        raise parse_results

    return parse_results

def is_missing(result):
    return result is None

class SELF(object):
    pass


class Evaluator(object):
    def __init__(self,  ns: Dict[str, Any], 
                        subst_context: typing.Optional[SubstitutionContext] = None, 
                        allow_unresolved: bool = False,
                        location: List[str] = []):
        self.ns = ns
        self.subst_context = subst_context
        self.location = location
        self.allow_unresolved = allow_unresolved

    def _resolve(self, value, in_formula=True, subst=True):
        if type(value) is str:
            if in_formula and value == "SELF":
                return SELF
            elif in_formula and value == "UNSET":
                return UNSET
            elif self.subst_context is not None and subst:
                try:
                    value = self.subst_context.evaluate(value, location=self.location)
                except (KeyError, AttributeError) as exc:
                    raise SubstitutionError(f"{value}: invalid key {exc}")
                except Exception as exc:
                    raise SubstitutionError(f"{value}: {exc}")
        return value

    def empty(self, *args, **kw):
        return ""

    def unset(self, *args, **kw):
        return UNSET

    def self_value(self, *args, **kw):
        return SELF

    def constant(self, value, subst=True, **kw):
        return self._resolve(value, subst=subst)

    def subexpression(self, value, subst=True):
        return self._evaluate_result(value, allow_unset=True, subst=subst)
    
    def namespace_lookup(self, *args, subst=True):
        if len(args) == 1 and type(args[0]) is ParseResults:
            args = args[0]
            assert args._name == "namespace_lookup"
        value = self.ns
        fields = list(args)
        while fields:
            fld = fields.pop(0)
            # check for wildcards
            if fld not in value and ('*' in fld or '?' in fld):
                names = sorted(fnmatch.filter(value.keys(), fld))
                if names:
                    fld = names[-1]
            # last element allowed to be UNSET, otherwise substitution error
            if fld not in value:
                if fields:
                    raise SubstitutionError(f"{'.'.join(self.location)}: '{fld}' undefined (in '{'.'.join(args)}')")
                else:
                    return UNSET('.'.join(args))
            # this can still throw an error if a nested interpolation is invoked
            try:
                value = value.get(fld, subst=subst)
            except (KeyError, AttributeError) as exc:
                raise SubstitutionError(f"{'.'.join(self.location)}: '{'.'.join(args)}' unresolved (at '{exc}')")
        return self._resolve(value, subst=subst)

    def _evaluate_result(self, parse_result, allow_unset=False, subst=True):
        allow_unset = allow_unset or self.allow_unresolved
        # if result is a handler, use evaluate
        if isinstance(parse_result, ResultsHandler):
            value = parse_result.evaluate(self)

        # if result is already a constant, resolve it
        elif type(parse_result) is not ParseResults:
            return self._resolve(parse_result, subst=subst)
        
        # lookup processing method based on name
        else:
            method = parse_result.getName()
            assert method is not None
            if not hasattr(self, method):
                raise ParserError(f"{'.'.join(self.location)}: don't know how to deal with an element of type '{method}'")

            try:
                value = getattr(self, method)(*parse_result, subst=subst)
            except SubstitutionError as exc:
                if allow_unset:
                    return UNSET("", [exc])
                raise

        if isinstance(value, Unresolved) and not allow_unset:
            raise UnsetError(f"'{value.value}' undefined", value.errors)

        return value

    def evaluate(self, value: Any, sublocation: List[str] = []) -> Any:
        """evaluates a single value, which can be a string with a formula and/or substitutions

        Args:
            value (Any): value to be evaluated. If not a str, returned as is
            sublocation (List[str]): location of value inside of nested container, as a list
                    of strings. Defaults to []

        Raises:
            ParserError: error parsing formula
            FormulaError: error evaluating formula
            SubstitutionError: unresolved symbol in formula or {}-substitution

        Returns:
            Any: result of evaluation
        """
        
        if type(value) is not str:
            return value

        loclen = len(self.location)
        self.location += sublocation

        try:
            if value.startswith("="):
                if value.startswith("=="):
                    return self._resolve(value[1:], in_formula=False)
                else:
                    try:
                        parse_results = parse_string(value[1:])
                    except Exception as exc:
                        raise ParserError(f"{'.'.join(self.location)}: error parsing formula '{value}'", exc)

                    try:
                        return self._evaluate_result(parse_results, allow_unset=True)
                    except Exception as exc:
                        raise FormulaError(f"{'.'.join(self.location)}: evaluation of '{value}' failed", exc, tb=True)
            else:
                try:
                    return self._resolve(value, in_formula=False)
                except Exception as exc:
                    raise SubstitutionError(f"{'.'.join(self.location)}: evaluation of '{value}' failed", exc)
        finally:
            self.location = self.location[:loclen]

    def evaluate_object(self, obj: Any,
                        sublocation: List[str] = [],
                        raise_substitution_errors: bool = True, 
                        recursion_level: int = 1,
                        verbose: bool = False) -> Any:
        """evaluates object, which can be a nested container, in which case string elements
        are evaluated recursively. Evaluations are done in place.

        Args:
            obj (Any): object to evaluate
            sublocation (List[str]): location of value inside of nested container, as a list
                    of strings. Defaults to []
            raise_substitution_errors (bool, optional): raise substitution errors, instead of returning
                Unresolved values. Defaults to True.
            recursion_level (int, optional): Limits recursion level into subcontainers. Defaults to 1,
                which means recurse once.
            verbose (bool, optional): Prints debug messages. Defaults to False.

        Raises:
            SubstitutionError: error in substitutions

        Returns:
            Any: object with evaluations
        """
        # string? evaluate directly and return
        if type(obj) is str:
            try:
                value = self.evaluate(obj, sublocation=sublocation)
                if isinstance(value, Unresolved) and (raise_substitution_errors or not self.allow_unresolved):
                    raise SubstitutionError(f"{'.'.join(sublocation)}: unresolved substitution", value.errors)
                return value
            except AttributeError as err:
                if raise_substitution_errors or not self.allow_unresolved:
                    raise SubstitutionError(f"{'.'.join(sublocation)}: substitution error", [err])
                return Unresolved(errors=[err])
            except SubstitutionError as err:
                if raise_substitution_errors or not self.allow_unresolved:
                    raise SubstitutionError(f"{'.'.join(sublocation)}: substitution error", [err])
                return Unresolved(errors=[err])
            
        # helper function
        def update(value, sloc):
            if isinstance(value, Unresolved):
                return value, False
            subloc = sublocation + [sloc]
            if verbose: 
                print(f"{subloc}: {value} ...")
            new_value = self.evaluate_object(value, raise_substitution_errors=raise_substitution_errors,
                                                recursion_level=recursion_level, verbose=verbose,
                                                sublocation=subloc)
            if verbose:
                print(f"{subloc}: {value} -> {new_value}")
            # UNSET return means delete or revert to default
            if new_value is UNSET:
                raise SubstitutionError(f"{'.'.join(self.location + subloc)}: UNSET not allowed here")
            # compare
            if isinstance(value, (dict, DictConfig, list, ListConfig)) or dataclasses.is_dataclass(value):
                updated = value is not new_value
            else:
                updated = value != new_value
            return new_value, updated

        obj_out = obj
        # recurse into containers?
        if recursion_level:
            recursion_level -= 1
            # use evaluate_dict() to recurse into dicts
            if isinstance(obj, (dict, DictConfig)):
                for key, value in obj.items():
                    new_value, value_updated = update(value, key)
                    new_key = self.evaluate(key, sublocation=sublocation)
                    if new_key != key:
                        value_updated = True
                    if value_updated:
                        if obj_out is obj:
                            obj_out = obj.copy()
                        if new_key != key:
                            del obj_out[key]
                            key = new_key
                        obj_out[key] = new_value
            # recurse into lists
            elif isinstance(obj, (list, ListConfig)):
                for i, value in enumerate(obj):
                    new_value, updated = update(value, f"#{i}")
                    if updated:
                        if obj_out is obj:
                            obj_out = obj.copy()
                        obj_out[i] = new_value
            # recurse into dataclasses
            elif dataclasses.is_dataclass(obj):
                newvals = {}
                for fld in dataclasses.fields(obj):
                    value = getattr(obj, fld.name)
                    new_value, updated = update(value, fld.name)
                    if updated:
                        newvals[fld.name] = new_value
                if newvals:
                    obj_out = dataclasses.replace(obj, **newvals)

        return obj_out


    def evaluate_dict(self, params: Dict[str, Any], 
                    corresponding_ns: typing.Optional[Dict[str, Any]] = None, 
                    defaults: Dict[str, Any] = {}, 
                    sublocation = [],
                    raise_substitution_errors: bool = True, 
                    collapse_substitution_errors: bool = False,  # true for subcontainers
                    subcontainer_type: typing.Optional[str] = None,
                    recursive: bool = True,
                    verbose: bool = False) -> Dict[str, Any]:
        """evaluates dict of parameters, which can contain nested containers which can be evaluated recursively.
        Returns new dict.

        Args:
            params (Dict[str, Any]): parameters to evaluate
            corresponding_ns (Optional[Dict[str, Any]], optional): corresponding namespace, into which evluations
                are propagated. Defaults to None.
            defaults (Dict[str, Any], optional): dictionary of defaults; UNSET values are cause defaults to
                be substituted in.
            sublocation (list, optional): location inside of nested container, as a list
                of strings. Defaults to []
            raise_substitution_errors (bool, optional): raises substitution errors. Defaults to True.
                if False, substitution errors are assigned to dict as Unresolved values.
            collapse_substitution_errors (bool, optional): if True, a substituion error inside the dict
                causes an Unresolved value to be returned instead of the dict. Defaults to False.
            subcontainer_type (str): type of subcontainer being evaluated. Defaults to None. Must be set
                if collapse_substitution_errors is True.
            recursive (bool, optional): recurse into subcontainers. Defaults to True.
            verbose (bool, optional): print debug messages. Defaults to False.

        Returns:
            Dict[str, Any]: _descopy of input dict with substitutions performed
        """
        if collapse_substitution_errors:
            assert subcontainer_type is not None
        params_out = params
        for name, value in list(params.items()):
            if isinstance(value, Unresolved):
                continue
            # 
            retry = True
            while retry:
                retry = False
                if verbose: # or type(value) is UNSET:
                    print(f"{name}: {value} ...")
                if type(value) is str:
                    try:
                        new_value = self.evaluate(value, sublocation=sublocation + [name])
                    except (AttributeError, SubstitutionError, ParserError, FormulaError) as err:
                        if raise_substitution_errors:
                            raise
                        new_value = Unresolved(errors=[str(err)])
                    if verbose:
                        print(f"{name}: {value} -> {new_value}")
                    # UNSET return means delete or revert to default
                    if new_value is UNSET:
                        if subcontainer_type:
                            raise SubstitutionError(f"{'.'.join(self.location + sublocation)}: UNSET not allowed here")
                        if params_out is params:
                            params_out = params.copy()
                        # if value is in defaults and is different, try to evaluate that instead
                        if name in defaults and defaults[name] is not UNSET and defaults[name] != value:
                            value = params_out[name] = defaults[name]
                            if corresponding_ns:
                                corresponding_ns[name] = defaults[name]
                            retry = True
                        else: 
                            if name in params_out:
                                del params_out[name]
                            if corresponding_ns and name in corresponding_ns:
                                del corresponding_ns[name]
                    elif new_value is not value and new_value != value:
                        if params_out is params:
                            params_out = OrderedDict(**params)
                        params_out[name] = new_value
                        if corresponding_ns:
                            corresponding_ns[name] = new_value
                elif isinstance(value, (dict, DictConfig)) and recursive:
                    value = self.evaluate_dict(
                        value,
                        corresponding_ns,
                        defaults,
                        sublocation=sublocation + [name],
                        raise_substitution_errors=raise_substitution_errors,
                        collapse_substitution_errors=True, subcontainer_type="Dict",
                        recursive=True,
                        verbose=verbose
                    )
                    params_out[name] = value
                    if corresponding_ns:
                        corresponding_ns[name] = value
                elif isinstance(value, (list, ListConfig)) and recursive:
                    # convert list to dict, and evaluate
                    proxy_dict = self.evaluate_dict(
                                {f"[{i}]": v for i, v in enumerate(value)},
                                corresponding_ns,
                                defaults,
                                sublocation=sublocation + [name],
                                raise_substitution_errors=raise_substitution_errors,
                                collapse_substitution_errors=True, subcontainer_type="List",
                                recursive=True,
                                verbose=verbose
                            )
                    if isinstance(proxy_dict, Unresolved):
                        value = proxy_dict
                    else:
                        value = type(value)(list(proxy_dict.values()))
                    params_out[name] = value
                    if corresponding_ns:
                        corresponding_ns[name] = value

        if collapse_substitution_errors:
            errors = []
            for elem in params_out.values():
                if isinstance(elem, Unresolved):
                    errors += elem.errors
            if errors:
                return Unresolved(f"unresolved {subcontainer_type} elements", errors)

        return params_out



if __name__ == "__main__":
    pass
